from typing import Dict, Optional

from pydantic import BaseModel


class RayConfig(BaseModel):
    address:str
    redis_password:str
    runtime_env: Dict


class TrainerConfig(BaseModel):
    backend:str
    num_workers:int
    use_gpu: bool


class LearningConfig(BaseModel):
    lr:float
    batch_size:int
    epochs: int
    steps_per_epoch: int



import ray
from ray.train import Trainer
import ray.train as train

from tensorflow.python.keras.callbacks import Callback
from raypipe.core.data_model import RayConfig, TrainerConfig


def start_ray(ray_config:RayConfig):
    try:
        ray.init(address=ray_config.address,
                 _redis_password=ray_config.redis_password,
                 runtime_env=ray_config.runtime_env
                 )
    except:
        ray.init()

def build_ray_trainer(trainer_cfg:TrainerConfig):
    return  Trainer(backend=trainer_cfg.backend,
                    num_workers=trainer_cfg.num_workers,
                    use_gpu=trainer_cfg.use_gpu)

class TrainReportCallback(Callback):
    def on_epoch_end(self, epoch, logs=None):
        train.report(**logs)


import logging  # 引入logging模块

class Logger(object):
    def __init__(self,name:str):
        logging.basicConfig(filename=name,level=logging.DEBUG,
                    format='%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s')

        self.logger=logging

    def info(self,*args):
        self.logger.info(*args)

    def warning(self,*args):
        msg, err = args
        self.logger.warning(msg, err)

    def debug(self, *args):
        self.logger.debug(*args)

    def error(self, *args):
        self.logger.debug(*args)


from typing import Dict
from raypipe.applications.base import BaseNN
import tensorflow as tf
import numpy as np


class CNN(BaseNN):
    def model_strategy_func(self,learning_cfg:Dict):
        learning_rate = learning_cfg.get("lr",0.001)

        model = tf.keras.Sequential([
            tf.keras.Input(shape=(28, 28)),
            tf.keras.layers.Reshape(target_shape=(28, 28, 1)),
            tf.keras.layers.Conv2D(32, 3, activation="relu"),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(128, activation="relu"),
            tf.keras.layers.Dense(10)
        ])

        model.compile(
            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
            optimizer=tf.keras.optimizers.SGD(learning_rate=learning_rate),
            metrics=["accuracy"])
        return model

    def data_generator_func(self,data_cfg:Dict,batch_size:int):
        (x_train, y_train), _ = tf.keras.datasets.mnist.load_data()
        # The `x` arrays are in uint8 and have values in the [0, 255] range.
        # You need to convert them to float32 with values in the [0, 1] range.
        x_train = x_train / np.float32(255)
        y_train = y_train.astype(np.int64)
        train_dataset = tf.data.Dataset.from_tensor_slices(
            (x_train, y_train)).shuffle(60000).repeat().batch(batch_size)

        return train_dataset





# This example showcases how to use Tensorflow with Ray Train.
# Original code:
# https://www.tensorflow.org/tutorials/distribute/multi_worker_with_keras
import argparse
import json
import os
import ray
import numpy as np

import ray.train as train
from ray.train import Trainer
import tensorflow as tf
from raypipe.core.rpipe.utils import TrainReportCallback


def mnist_dataset(batch_size):
    (x_train, y_train), _ = tf.keras.datasets.mnist.load_data()
    # The `x` arrays are in uint8 and have values in the [0, 255] range.
    # You need to convert them to float32 with values in the [0, 1] range.
    x_train = x_train / np.float32(255)
    y_train = y_train.astype(np.int64)
    train_dataset = tf.data.Dataset.from_tensor_slices(
        (x_train, y_train)).shuffle(60000).repeat().batch(batch_size)
    return train_dataset


def build_and_compile_cnn_model(config):
    learning_rate = config.get("lr", 0.001)
    model = tf.keras.Sequential([
        tf.keras.Input(shape=(28, 28)),
        tf.keras.layers.Reshape(target_shape=(28, 28, 1)),
        tf.keras.layers.Conv2D(32, 3, activation="relu"),
        tf.keras.layers.Flatten(),
        tf.keras.layers.Dense(128, activation="relu"),
        tf.keras.layers.Dense(10)
    ])
    model.compile(
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=tf.keras.optimizers.SGD(learning_rate=learning_rate),
        metrics=["accuracy"])
    return model


def train_func(config):
    per_worker_batch_size = config.get("batch_size", 64)
    epochs = config.get("epochs", 3)
    steps_per_epoch = config.get("steps_per_epoch", 70)

    tf_config = json.loads(os.environ["TF_CONFIG"])
    num_workers = len(tf_config["cluster"]["worker"])

    strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy()

    global_batch_size = per_worker_batch_size * num_workers
    multi_worker_dataset = mnist_dataset(global_batch_size)

    with strategy.scope():
        # Model building/compiling need to be within `strategy.scope()`.
        multi_worker_model = build_and_compile_cnn_model(config)

    history = multi_worker_model.fit(
        multi_worker_dataset,
        epochs=epochs,
        steps_per_epoch=steps_per_epoch,
        callbacks=[TrainReportCallback()])
    results = history.history
    return results


def train_tensorflow_mnist(num_workers=2, use_gpu=False, epochs=4):
    trainer = Trainer(
        backend="tensorflow", num_workers=num_workers, use_gpu=use_gpu)
    trainer.start()
    results = trainer.run(
        train_func=train_func,
        config={
            "lr": 1e-3,
            "batch_size": 64,
            "epochs": epochs
        })
    trainer.shutdown()
    print(f"Results: {results[0]}")


if __name__ == "__main__":
    # runtime_env = {"container":{"image": "docker.4pd.io/ray_tensorflow2:pipe-105-commit-aa96d082"}}
    # runtime_env = {"pip": [ "tensorflow-cpu==2.3.0"]}
    ray.init(address="ray://172.27.69.28:32071",
             _redis_password='5241590000000000')
             # runtime_env=runtime_env)

    train_tensorflow_mnist(
        num_workers=1,
        use_gpu=False,
        epochs=2)


import unittest

from raypipe.applications.cnn import CNN


class CNNTester(unittest.TestCase):
    def test_cnn_remote_train(self):
        general_cfg={
            "ray_cfg":{
                "address":"ray://172.27.69.28:32071",
                "redis_password" : '5241590000000000',
                "runtime_env":{}
            },
            "trainer_cfg":{
                "backend":"tensorflow",
                "num_workers":1,
                "use_gpu":False
            },
            "learning_cfg":{
                "lr":1e-3,
                "batch_size": 10,
                "epochs": 3,
                "steps_per_epoch": 3
            }
        }
        cnn=CNN(**general_cfg)
        cnn.train(method="remote")


import ray

from raypipe import logger
from raypipe.core.data_model import RayConfig, TrainerConfig, LearningConfig
from raypipe.core.rpipe.utils import start_ray


def init(cls):
    default_ray_cfg = {"address": "auto",
                       "redis_password": 5241590000000000,
                       "runtime_env": {}}
    default_trainer_cfg = {"backend": "tensorflow", "num_workers": 1, "use_gpu": False}
    default_learning_cfg = {"lr": 1e-5, "batch_size": 16, "epochs": 3, "steps_per_epoch": 10}

    def wrapper(*args, **general_cfg):
        ray_cfg = general_cfg.get("ray_cfg")
        trainer_cfg = general_cfg.get("trainer_cfg")
        learning_cfg = general_cfg.get("learning_cfg")
        data_cfg=general_cfg.get("data_cfg",{})

        if not ray_cfg:
            start_ray(RayConfig(**default_ray_cfg))

        if not ray.is_initialized() and ray_cfg:
            start_ray(RayConfig(**ray_cfg))
        logger.info('=========== Ray engine prepared =========== ')

        if not trainer_cfg:
            trainer_config = TrainerConfig(**default_trainer_cfg)
        else:
            trainer_config = TrainerConfig(**trainer_cfg)

        if not learning_cfg:
            learning_config = LearningConfig(**default_learning_cfg)
        else:
            learning_config = LearningConfig(**learning_cfg)

        cls._trainer_config = trainer_config
        cls._learning_config = learning_config
        cls._data_cfg=data_cfg

        logger.info('=========== %s initialized=========== ' % cls.__name__)
        return cls

    return wrapper


def is_init(func):
    def wrapper(self):
        if not ray.is_initialized:
            raise EnvironmentError("Ray session not initialized")
        return func

    return wrapper



from raypipe.core.model_proxy import ModelProxy


class BaseNN(object):
    def model_strategy_func(self):
        raise NotImplementedError("Please implement model strategy")

    def data_generator_func(self):
        raise NotImplementedError("Please implement data generator")

    def __init__(self, **general_cfg):
        if "ray_cfg" not in general_cfg:
            raise not NotImplementedError("Please configure ray_cfg")
        self._ray_cfg=general_cfg.get("ray_cfg")

        if "trainer_cfg" not in general_cfg:
            raise not NotImplementedError("Please configure trainer_cfg")
        self._trainer_cfg=general_cfg.get("trainer_cfg")

        if "learning_cfg" not in general_cfg:
            raise not NotImplementedError("Please configure learning_cfg")
        self._learning_cfg = general_cfg.get("learning_cfg")

        self._model_proxy=ModelProxy(**general_cfg)\
            (self.model_strategy_func,self.data_generator_func)

    def train(self,method="remote"):
        if method=="remote":
            self._model_proxy.submit()
        else:
            self._model_proxy.local_train()

    def serve(self,method="remote"):
        pass





from raypipe import logger
from raypipe.core import rpipe
import tensorflow as tf

from raypipe.core.data_model import LearningConfig
from raypipe.core.rpipe.utils import build_ray_trainer, TrainReportCallback


@rpipe.init
class ModelProxy:
    def __init__(self, model_strategy_func,data_generator_func):
        """
        paras as model config
        :param kwargs:
        """
        self._validate_model(model_strategy_func)
        self._validate_learning_config(self._learning_config)

        self.model_strategy_func = model_strategy_func
        self.data_generator_func=data_generator_func

        self.ray_trainer = build_ray_trainer(self._trainer_config)

        if not self.ray_trainer: raise NotImplementedError("Trainer not built.")
        self.initialized=True

    @rpipe.is_init
    def _validate_env(self):
        if not self.initialized:
            raise EnvironmentError("Model distribution not initialized")

    def _validate_model(self,model):
        #todo
        pass

    def _validate_learning_config(self, config:LearningConfig):
        #todo
        pass

    def local_train(self,dataset):
        self._validate_env()
        local_model=self.model_strategy_func(self.learning_cfg)
        local_model.fit(dataset)

    def local_eval(self):
        pass

    def _train_template(self,general_config):
        learning_config=general_config.get("learning_config")
        trainer_config = general_config.get("trainer_config")
        data_cfg= general_config.get("data_cfg")

        global_batch_size = learning_config.batch_size * trainer_config.num_workers
        batch_dataset = self.data_generator_func(data_cfg,global_batch_size)

        strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy()
        with strategy.scope():
            multi_worker_model = self.model_strategy_func(learning_config.json())

        history = multi_worker_model.fit(
            batch_dataset,
            epochs=self.learning_config.epochs,
            steps_per_epoch=self.learning_config.steps_per_epoch,
            callbacks=[TrainReportCallback()])
        results = history.history
        return results

    def submit(self):
        """
        :return:
        """
        self._validate_env()
        self.ray_trainer.start()
        logger.info('=========== Trainer started =========== ')

        self.ray_trainer.run(
            train_func=self._train_template,
            config={
                    "learning_config":self._learning_config,
                    "trainer_config":self._trainer_config,
                    "data_cfg":self._data_cfg}
            )

    def upload(self):
        """
        :return:
        """
        self._validate_env()


    def collect(self):
        """
        :return:
        """
        self._validate_env()

    def shutdown(self):
        self.ray_trainer.shutdown()
        self.ray_trainer=None
        logger.info('=========== Model training ended =========== ')






import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raypipe",
    version="1.0.0",
    author="Billy Yan",
    author_email="",
    description="Easy implementation and abstraction for train and deploy model on ray cluster",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/billyyann/RayPipe",
    packages=setuptools.find_packages(),
    install_requires=['tensorflow-cpu==2.3.0'],
    entry_points={
        'console_scripts': [
        ],
    },
    setup_requires=[
        # 'pytest-runner==3.0',
        # setuptools需要使用50.0.0以上版本，否则执行pytest的时候，无法从我司自己的pypi里下载
        'setuptools>=50.0.0',
        'ruamel.yaml',
        'wheel'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)